@@ -64,16 +64,10 @@ name.check.lsd <- function( file, col.names = NULL, check.names = TRUE ){
 info.dimensions.lsd <- function( file ){
 
   # read from disk
-  dataSet <- utils::read.delim( file, na.strings = "NA" )
-
-  if( nrow( dataSet ) == 0 )            # invalid file?
-    stop( paste0( "File '", file, "' is invalid!") )
-
-  if( all( is.na( dataSet[ , ncol( dataSet ) ]) ) )  # remove empty last column
-    dataSet <- dataSet[ , - ncol( dataSet )]
+  dataSet <- read.raw.lsd( file )
 
   # caclulate statistics
-  tSteps <- nrow( dataSet ) - 1
+  tSteps <- nrow( dataSet )
   nVars <- ncol( dataSet )
   varNames <- name.clean.lsd( colnames( dataSet ) )
 
@@ -85,18 +79,16 @@ info.dimensions.lsd <- function( file ){
 
 info.names.lsd <- function( file ){
 
-  # read from disk
-  dataSet <- utils::read.delim( file, na.strings = "NA", nrows = 1,
-                                stringsAsFactors = FALSE )
+  # read header line (labels) from disk
+  header <- scan( file, what = character( ), sep = "\t", quote = NULL,
+                  nlines = 1, quiet = TRUE )
+  header <- header[ 1 : ( length( header ) - 1 ) ]  # remove last tab
 
-  if( nrow( dataSet ) == 0 )            # invalid file?
+  if( length( header ) == 0 )            # invalid file?
     stop( paste0( "File '", file, "' is invalid!") )
 
-  if( all( is.na( dataSet[ , ncol( dataSet ) ]) ) )  # remove empty last column
-    dataSet <- dataSet[ , - ncol( dataSet )]
-
   # extract labels and remove duplicates
-  lsd.name <- unique( name.var.lsd( colnames( dataSet ) ) )
+  lsd.name <- unique( name.var.lsd( make.names( header ) ) )
 
   return( lsd.name )
 }
@@ -107,20 +99,7 @@ info.names.lsd <- function( file ){
 info.init.lsd <- function( file ){
 
   # read from disk
-  dataSet <- as.matrix( utils::read.delim( file, na.strings = "NA", nrows = 1,
-                                           colClasses = "numeric",
-                                           stringsAsFactors = FALSE ) )
-
-  if( nrow( dataSet ) == 0 )            # invalid file?
-    stop( paste0( "File '", file, "' is invalid!") )
-
-  # remove empty last column
-  if( all( is.na( dataSet[ , ncol( dataSet ) ] ) ) )
-    dataSet <- as.matrix( t( dataSet[ , - ncol( dataSet ) ] ) )
-
-  # reformat labels
-  colnames( dataSet ) <- name.clean.lsd( colnames( dataSet ) )
-  rownames( dataSet ) <- "0"
+  dataSet <- read.raw.lsd( file, nrows = 0 )
 
   return( dataSet )
 }
@@ -131,18 +110,7 @@ info.init.lsd <- function( file ){
 info.details.lsd <- function( file ){
 
   # read from disk
-  dataSet <- utils::read.delim( file, na.strings = "NA", nrows = 1,
-                                stringsAsFactors = FALSE )
-
-  if( nrow( dataSet ) == 0 )            # invalid file?
-    stop( paste0( "File '", file, "' is invalid!") )
-
-  # remove empty last column
-  if( all( is.na( dataSet[ , ncol( dataSet ) ]) ) )
-    dataSet <- dataSet[ , - ncol( dataSet )]
-
-  # reformat column labels
-  colnames( dataSet ) <- name.clean.lsd( colnames( dataSet ) )
+  dataSet <- read.raw.lsd( file, nrows = 0 )
 
   # get the most "deep" object position
   maxPosit <- 1
@@ -193,7 +161,9 @@ info.details.lsd <- function( file ){
 
 # ==== Compute statistics from multiple runs ====
 
-info.stats.lsd <- function( array, rows = 1, cols = 2 ){
+info.stats.lsd <- function( array, rows = 1, cols = 2, median = FALSE,
+                            ci = "none", ci.conf = 0.95, ci.boot = NULL,
+                            boot.R = 999, na.rm = TRUE, inf.rm = TRUE ) {
 
   # Get dimension data
   dimArray <- dim( array )
@@ -219,24 +189,48 @@ info.stats.lsd <- function( array, rows = 1, cols = 2 ){
     array( as.numeric( NA ), dim = c( dimArray[ rows ], dimArray[ cols ] ),
            dimnames = list( dimNames[[ rows ]], dimNames[[ cols ]] ) )
 
-  # prepare mask for dimension selection
-  baseMask <- list( )
-  for( k in 1 : nDimArray ){
-    if( rows == k || cols == k )        # dimensions to show
-      baseMask[[ k ]] <- rep( FALSE, dimArray[ k ])
-    else
-      baseMask[[ k ]] <- rep( TRUE, dimArray[ k ])
+  if( median )
+    med <- avg
+
+  if( ! ci %in% c( "none", "mean", "median", "auto" ) ) {
+    ci <- "none"
+    warning( "Invalid confidence interval statistic, ignoring" )
   }
 
-  # Compute averages, std. deviation etc. and store in 2D arrrays
-  for( j in 1 : dimArray[ cols ] )
-    for( i in 1 : dimArray[ rows ] ){
+  if( ci == "auto" ) {
+    if( median )
+      ci <- "median"
+    else
+      ci <- "mean"
+  }
+
+  if( ! is.null( ci.boot ) &&
+      ! ci.boot %in% c( "basic", "perc", "bca" ) ) {
+    ci.boot <- NULL
+    warning( "Invalid bootstrap confidence interval type, ignoring" )
+  }
+
+  if( ci != "none" )
+    ci.lo <- ci.hi <- avg
+
+  # prepare mask for dimension selection
+  baseMask <- list( )
+  for( k in 1 : nDimArray ) {
+    if( rows == k || cols == k )        # dimensions to show
+      baseMask[[ k ]] <- rep( FALSE, dimArray[ k ] )
+    else
+      baseMask[[ k ]] <- rep( TRUE, dimArray[ k ] )
+  }
+
+  # Compute averages, std. deviation etc. and store in 2D arrays
+  for( j in 1 : dimArray[ cols ] ) {
+    for( i in 1 : dimArray[ rows ] ) {
 
       # Get the appropriate vector (3D array) or matrix (4D) for analysis
       first <- TRUE
       mask <- baseMask
       for( k in 1 : nDimArray )       # adjust the mask for (i,j)
-        if( rows == k || cols == k ){
+        if( rows == k || cols == k ) {
           if( first ){
             mask[[ k ]][ i ] <- TRUE
             first <- FALSE
@@ -250,24 +244,100 @@ info.stats.lsd <- function( array, rows = 1, cols = 2 ){
         elem <- array[ mask[[ 1 ]], mask[[ 2 ]], mask[[ 3 ]], mask[[ 4 ]] ]
 
       # calculate the statistics
-      avg[ i, j ] <- mean( elem, na.rm = TRUE )
-      sDev[ i, j ] <- stats::sd( elem, na.rm = TRUE )
-      if( is.finite( avg[ i, j ] ) ){
-        M[ i, j ] <- max( elem, na.rm = TRUE )
-        m[ i, j ] <- min( elem, na.rm = TRUE )
+      if( na.rm )
+        elem <- elem[ ! is.na( elem ) ]
+
+      if( inf.rm )
+        elem <- elem[ is.finite( elem ) ]
+
+      n = length( elem[ is.finite( elem ) ] )
+      avg[ i, j ] <- mean( elem )
+      sDev[ i, j ] <- stats::sd( elem )
+
+      if( n > 0 ) {
+        M[ i, j ] <- max( elem )
+        m[ i, j ] <- min( elem )
       }
-      else{                           # avoid Inf/-Inf when all is NA
-		avg[ i, j ] <- NA
-		sDev[ i, j ] <- NA
+      else {                           # avoid Inf/-Inf when all is NA
         M[ i, j ] <- NA
         m[ i, j ] <- NA
       }
-    }
 
-  if( ! transp )
-    return( list( avg = avg, sd = sDev, max = M, min = m ) )
-  else
-    return( list( avg = t( avg ), sd = t( sDev ), max = t( M ), min = t( m ) ) )
+      if( median )
+        med[ i, j ] <- stats::median( elem )
+
+      if( ci != "none" ) {
+        ci.lo[ i, j ] <- ci.hi[ i, j ] <- NA
+
+        if( n > 0 ) {
+          if( is.null( ci.boot ) ) {
+            if( ci == "mean" ) {
+              d = abs( stats::qt( ( 1 - ci.conf ) / 2, n - 1 ) ) * sDev[ i, j ] / sqrt( n )
+
+              if( is.finite( avg[ i, j ] ) && ! is.null( d ) && is.finite( d ) ) {
+                ci.lo[ i, j ] = avg[ i, j ] - d
+                ci.hi[ i, j ] = avg[ i, j ] + d
+              }
+            } else {
+              c <- NULL
+              try( c <- suppressWarnings( stats::wilcox.test( elem, conf.int = TRUE,
+                                                              conf.level = ci.conf,
+                                                              digits.rank = 7 )$conf.int ),
+                   silent = TRUE )
+              if( ! is.null( c ) && is.finite( c[ 1 ] ) )
+                ci.lo[ i, j ] <- c[ 1 ]
+
+              if( ! is.null( c ) && is.finite( c[ 2 ] ) )
+                ci.hi[ i, j ] <- c[ 2 ]
+            }
+          } else {
+            if( ci == "mean" )
+              f <- function( data, sel ) mean( data[ sel ] )
+            else
+              f <- function( data, sel ) stats::median( data[ sel ] )
+
+            b <- c <- NULL
+            try( invisible( utils::capture.output( b <- boot::boot( elem,
+                                                                    statistic = f,
+                                                                    R = boot.R ) ) ),
+                 silent = TRUE )
+
+            if( ! is.null( b ) ) {
+              try( invisible( utils::capture.output( c <- boot::boot.ci( b,
+                                                                         conf = ci.conf,
+                                                                         type = ci.boot ) ) ),
+                   silent = TRUE )
+            }
+
+            if( ci.boot == "perc" )
+              ci.boot = "percent"       # adjust name difference in data
+
+            if( ! is.null( c[[ ci.boot ]] ) && is.finite( c[[ ci.boot ]][ 4 ] ) )
+              ci.lo[ i, j ] <- c[[ ci.boot ]][ 4 ]
+
+            if( ! is.null( c[[ ci.boot ]] ) && is.finite( c[[ ci.boot ]][ 5 ] ) )
+              ci.hi[ i, j ] <- c[[ ci.boot ]][ 5 ]
+          }
+        }
+      }
+    }
+  }
+
+  res <- list( avg = avg, sd = sDev, max = M, min = m )
+
+  if( median )
+    res[[ "med" ]] <- med
+
+  if( ci != "none" ) {
+    res[[ "ci.lo" ]] <- ci.lo
+    res[[ "ci.hi" ]] <- ci.hi
+  }
+
+  if( transp )
+    for( i in 1 : length( res ) )
+      res[[ i ]] = t( res[[ i ]] )
+
+  return( res )
 }
 
 
@@ -308,7 +378,7 @@ select.colnames.lsd <- function( dataSet, col.names, instance = 0 ){
 
 # ==== Select a subset of a data frame (by variable attributes) ====
 
-select.colattrs.lsd <- function( dataSet, info, col.names = NA, posit = NULL,
+select.colattrs.lsd <- function( dataSet, info, col.names = NULL, posit = NULL,
                                  init.value = NA, init.time = NA, end.time = NA ){
 
   # test if files are compatible (in principle)
@@ -317,7 +387,8 @@ select.colattrs.lsd <- function( dataSet, info, col.names = NA, posit = NULL,
     stop( "Info table invalid or incompatible with provided dataSet" )
 
   # format valid names for matching
-  col.names <- make.names( name.clean.lsd( col.names ) )
+  if( ! is.null( col.names ) )
+    col.names <- make.names( name.clean.lsd( col.names ) )
 
   # check if position is not formatted as text and convert if needed
   if( ! is.null( posit ) && ! is.character( posit ) ){
@@ -336,7 +407,7 @@ select.colattrs.lsd <- function( dataSet, info, col.names = NA, posit = NULL,
   for( i in 1 : nrow( info ) ){
 
     # if column names are specified, check if belongs to the set
-    if( ! is.na( col.names ) && ! ( info$R_name[ i ] %in% col.names ) )
+    if( ! is.null( col.names ) && ! ( info$R_name[ i ] %in% col.names ) )
       next
 
     # check if value attributes match
@@ -378,31 +449,41 @@ select.colattrs.lsd <- function( dataSet, info, col.names = NA, posit = NULL,
 # ==== Read LSD results file and clean variables names ====
 
 read.raw.lsd <- function( file, nrows = -1, skip = 0 ){
-  # read from disk
-  dataSet <- as.matrix( utils::read.delim( file, na.strings = "NA",
-                                           colClasses = "numeric",
-                                           stringsAsFactors = FALSE ) )
+
+  # read header line (labels) from disk
+  header <- scan( file, what = character( ), sep = "\t", quote = NULL,
+                  nlines = 1, quiet = TRUE )
+  header <- header[ 1 : ( length( header ) - 1 ) ]  # remove last tab
+  header <- make.names( header )
+
+  if( length( header ) == 0 )            # invalid file?
+    stop( paste0( "File '", file, "' is invalid!") )
+
+  # try to calculate data size
+  if ( nrows > 0 ) {
+    n <- length( header ) * ( nrows - skip )
+  } else {
+    if ( nrows < 0 ) {
+      n <- -1
+    } else {                            # nrows = 0 : get initial values
+      n <- length( header )
+      skip <- -1
+    }
+  }
+
+  # read data from disk
+  dataSet <- matrix( scan( file, what = numeric( ), n = n, quote = NULL,
+                           skip = skip + 2, na.strings = "NA", quiet = TRUE ),
+                     ncol = length( header ), byrow = TRUE )
 
   if( nrow( dataSet ) == 0 )            # invalid file?
     stop( paste0( "File '", file, "' is invalid!") )
 
-  # remove unwanted rows and columns (artifacts)
-  dataSet <- dataSet[ -1, ]             # remove first data row (initial conditions)
-  if( all( is.na( dataSet[ , ncol( dataSet ) ]) ) )  # remove empty last column
-    dataSet <- dataSet[ , - ncol( dataSet )]
-
-  # remove rows to discard (user)
-  if( skip > 0 )                        # there are rows to skip?
-    dataSet <- dataSet[ - c( 1 : skip ), ]
-  if( nrows >= 0 && nrow( dataSet ) >= nrows) # there are rows to remove?
-    dataSet <- dataSet[ c( 0 : nrows ), ]
-
   # adjust rown labels
-  if( nrow( dataSet ) > 0 )             #any row to label?
-    rownames( dataSet ) <- c( ( 1 + skip ) : ( nrow( dataSet ) + skip ) )
+  rownames( dataSet ) <- c( ( 1 + skip ) : ( nrow( dataSet ) + skip ) )
 
   # reformat column labels
-  colnames( dataSet ) <- name.clean.lsd( colnames( dataSet ) )
+  colnames( dataSet ) <- name.clean.lsd( header )
 
   return( dataSet )
 }
